import java.util.*;

public class Problem060 {

	private static int MAX_SMALL = 10000;
	private static int MAX_FIFTH = 100000;
	private static int MAX_PRIME = 1000000;

    /**
     * Efficient primality test based on the Miller-Rabin algorithm.
     * This is needed since Java doesn't have a built-in primality tester like sympy.
     */
    public static boolean isPrime(long n) {
        // Handle basic cases
        if (n <= 1) return false;
        if (n <= 3) return true;
        if (n % 2 == 0 || n % 3 == 0) return false;
        
        // Check using 6k +/- 1 pattern
        long i = 5;
        while (i * i <= n) {
            if (n % i == 0 || n % (i + 2) == 0) {
                return false;
            }
            i += 6;
        }
        return true;
    }
    
    /**
     * Concatenate two numbers by converting them to strings.
     */
    public static long concatenate(int a, int b) {
        return Long.parseLong(a + "" + b);
    }
    
    /**
     * Check if a list of primes satisfies the property that
     * any two primes concatenated (in either order) form a prime.
     */
    public static boolean checkPrimePairProperty(List<Integer> primes) {
        for (int i = 0; i < primes.size(); i++) {
            for (int j = i + 1; j < primes.size(); j++) {
                int p1 = primes.get(i);
                int p2 = primes.get(j);
                
                if (!isPrime(concatenate(p1, p2)) || !isPrime(concatenate(p2, p1))) {
                    return false;
                }
            }
        }
        return true;
    }
    
    /**
     * Find primes that are compatible with the existing set.
     */
    public static List<Integer> findCompatiblePrimes(List<Integer> existingSet, List<Integer> candidatePrimes) {
        List<Integer> compatiblePrimes = new ArrayList<>();
        
        for (int prime : candidatePrimes) {
            if (existingSet.contains(prime)) {
                continue;
            }
            
            boolean compatible = true;
            for (int existingPrime : existingSet) {
                if (!isPrime(concatenate(prime, existingPrime)) || !isPrime(concatenate(existingPrime, prime))) {
                    compatible = false;
                    break;
                }
            }
            
            if (compatible) {
                compatiblePrimes.add(prime);
            }
        }
        
        return compatiblePrimes;
    }
    
    /**
     * Generate a list of prime numbers up to maxPrime.
     * This is a simple implementation of the Sieve of Eratosthenes.
     */
    public static List<Integer> generatePrimes(int maxPrime) {
        boolean[] isComposite = new boolean[maxPrime + 1];
        List<Integer> primes = new ArrayList<>();
        
        for (int i = 2; i <= maxPrime; i++) {
            if (!isComposite[i]) {
                primes.add(i);
                
                // Mark multiples as composite
                long j = (long)i * i;
                while (j <= maxPrime) {
                    isComposite[(int)j] = true;
                    j += i;
                }
            }
        }
        
        return primes;
    }
    
    /**
     * Filter a list of primes to only include those less than a certain threshold.
     */
    public static List<Integer> filterPrimesLessThan(List<Integer> primes, int threshold) {
        List<Integer> filteredPrimes = new ArrayList<>();
        for (int prime : primes) {
            if (prime < threshold) {
                filteredPrimes.add(prime);
            }
        }
        return filteredPrimes;
    }
    
    /**
     * Calculate the sum of a list of integers.
     */
    public static int calculateSum(List<Integer> numbers) {
        int sum = 0;
        for (int number : numbers) {
            sum += number;
        }
        return sum;
    }
    
    /**
     * Find the set of 'size' primes with the lowest sum where
     * any two primes concatenated form a prime.
     * Returns a Map with the best set and its sum.
     */
    public static Map<String, Object> findLowestSumSet(int size, int maxPrime) {
        // Generate prime numbers up to maxPrime
        List<Integer> primes = generatePrimes(maxPrime);
        
        // Start with the known set of four primes
        List<Integer> knownSet = Arrays.asList(3, 7, 109, 673);
        
        Map<String, Object> result = new HashMap<>();
        
        if (size == 4) {
            result.put("set", knownSet);
            result.put("sum", calculateSum(knownSet));
            return result;
        }
        
        // Find compatible primes to add to the known set
        List<Integer> compatiblePrimes = findCompatiblePrimes(knownSet, primes);
        Collections.sort(compatiblePrimes);  // Sort by value to prioritize smaller primes
        
        if (!compatiblePrimes.isEmpty()) {
            int bestPrime = compatiblePrimes.get(0);
            List<Integer> bestSet = new ArrayList<>(knownSet);
            bestSet.add(bestPrime);
            
            result.put("set", bestSet);
            result.put("sum", calculateSum(bestSet));
            return result;
        }
        
        // If no compatible prime found, try a different approach:
        // Look for sets of 5 primes from scratch
        int bestSum = Integer.MAX_VALUE;
        List<Integer> bestSet = new ArrayList<>();
        
        // Start with small primes for better performance
        List<Integer> smallPrimes = filterPrimesLessThan(primes, MAX_SMALL);
        
        // Try to build sets directly with the smallest primes
        for (int i = 0; i < Math.min(30, smallPrimes.size()); i++) {
            int p1 = smallPrimes.get(i);
            List<Integer> set1 = Collections.singletonList(p1);
            List<Integer> compatible1 = findCompatiblePrimes(set1, smallPrimes);
            
            for (int p2 : compatible1) {
                if (p2 <= p1) {
                    continue;  // Skip to avoid duplicates
                }
                
                List<Integer> set2 = Arrays.asList(p1, p2);
                List<Integer> compatible2 = findCompatiblePrimes(set2, smallPrimes);
                
                for (int p3 : compatible2) {
                    if (p3 <= p2) {
                        continue;
                    }
                    
                    List<Integer> set3 = Arrays.asList(p1, p2, p3);
                    List<Integer> compatible3 = findCompatiblePrimes(set3, smallPrimes);
                    
                    for (int p4 : compatible3) {
                        if (p4 <= p3) {
                            continue;
                        }
                        
                        List<Integer> set4 = Arrays.asList(p1, p2, p3, p4);
                        List<Integer> compatible4 = findCompatiblePrimes(set4, smallPrimes);
                        
                        for (int p5 : compatible4) {
                            if (p5 <= p4) {
                                continue;
                            }
                            
                            // Need to create a mutable list to properly verify
                            List<Integer> candidateSet = new ArrayList<>();
                            candidateSet.add(p1);
                            candidateSet.add(p2);
                            candidateSet.add(p3);
                            candidateSet.add(p4);
                            candidateSet.add(p5);
                            
                            int candidateSum = calculateSum(candidateSet);
                            
                            if (candidateSum < bestSum) {
                                // Verify the whole set
                                if (checkPrimePairProperty(candidateSet)) {
                                    bestSum = candidateSum;
                                    bestSet = new ArrayList<>(candidateSet);
                                    System.out.println("Found better set: " + bestSet + " with sum " + bestSum);
                                }
                            }
                        }
                    }
                }
            }
        }
        
        result.put("set", bestSet);
        result.put("sum", bestSum);
        return result;
    }
    
    /**
     * Search for a fifth prime that's compatible with the known set of four primes.
     * Uses a more extensive search than the main function.
     */
    public static Map<String, Object> searchForCompatibleFifthPrime(List<Integer> knownSet, int maxPrime) {
        // Generate more prime numbers to test
        List<Integer> candidatePrimes = generatePrimes(maxPrime);
        
        List<Integer> compatiblePrimes = findCompatiblePrimes(knownSet, candidatePrimes);
        Collections.sort(compatiblePrimes);  // Sort by value
        
        Map<String, Object> result = new HashMap<>();
        
        if (!compatiblePrimes.isEmpty()) {
            int bestPrime = compatiblePrimes.get(0);
            List<Integer> bestSet = new ArrayList<>(knownSet);
            bestSet.add(bestPrime);
            
            result.put("set", bestSet);
            result.put("sum", calculateSum(bestSet));
            return result;
        }
        
        result.put("set", new ArrayList<>());
        result.put("sum", 0);
        return result;
    }
    
    /**
     * Create different combinations of immutable Lists for our algorithms.
     * This is a utility function to help with creating different sets.
     */
    public static List<Integer> createList(int... values) {
        List<Integer> list = new ArrayList<>();
        for (int value : values) {
            list.add(value);
        }
        return list;
    }
    
    /**
     * Verify all concatenations of prime pairs are prime
     */
    public static void verifyPrimePairs(List<Integer> primes) {
        for (int i = 0; i < primes.size(); i++) {
            for (int j = i + 1; j < primes.size(); j++) {
                int p1 = primes.get(i);
                int p2 = primes.get(j);
                
                long c1 = concatenate(p1, p2);
                long c2 = concatenate(p2, p1);
                
                System.out.println(p1 + " + " + p2 + " = " + c1 + " (Prime: " + isPrime(c1) + ")");
                System.out.println(p2 + " + " + p1 + " = " + c2 + " (Prime: " + isPrime(c2) + ")");
            }
        }
    }
    
    public static void main(String[] args) {
        System.out.println("Starting search for a set of five primes...");
        
        // First try with the known set of four primes
        List<Integer> knownSet = Arrays.asList(3, 7, 109, 673);
        Map<String, Object> result = searchForCompatibleFifthPrime(knownSet, MAX_FIFTH);
        
        List<Integer> resultSet = (List<Integer>) result.get("set");
        int sumValue = (Integer) result.get("sum");
        
        if (resultSet != null && !resultSet.isEmpty()) {
            System.out.println("Found solution: " + resultSet);
            System.out.println("Sum: " + sumValue);
            
            // Verify all concatenations are prime
            verifyPrimePairs(resultSet);
        } else {
            System.out.println("No solution found extending the known set. Trying from scratch...");
            
            // Try a more general search
            result = findLowestSumSet(5, MAX_PRIME);
            
            resultSet = (List<Integer>) result.get("set");
            sumValue = (Integer) result.get("sum");
            
            if (resultSet != null && !resultSet.isEmpty()) {
                System.out.println("Found solution: " + resultSet);
                System.out.println("Sum: " + sumValue);
                
                // Verify all concatenations are prime
                verifyPrimePairs(resultSet);
            } else {
                System.out.println("No solution found. Try increasing the search range.");
            }
        }
    }
}
