import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;

public class Problem345 {

    // Method to calculate the maximum sum using the Hungarian Algorithm
    public static double calculateMaxMatrixSum(double[][] matrix) {
        int N = matrix.length;
        if (N == 0 || matrix[0].length != N) {
            throw new IllegalArgumentException("Matrix must be square and non-empty.");
        }

        // 1. Create the cost matrix (negated values)
        double[][] costMatrix = new double[N][N];
        for (int i = 0; i < N; i++) {
            for (int j = 0; j < N; j++) {
                costMatrix[i][j] = -matrix[i][j];
            }
        }

        // 2. Solve the assignment problem (minimization on cost matrix)
        // This part requires an actual Hungarian Algorithm implementation.
        // Let's assume we have a class 'HungarianSolver'
        HungarianSolver solver = new HungarianSolver(costMatrix);
        int[] assignment = solver.execute(); // Returns assignment: assignment[row] = col

        // 3. Calculate the minimum cost from the assignment
        double minCost = 0;
        for (int i = 0; i < N; i++) {
            if (assignment[i] != -1) { // Check if row i has an assignment
                 minCost += costMatrix[i][assignment[i]];
            } else {
                 // Handle case where no assignment is found (should not happen for square matrix if algo is correct)
                 System.err.println("Warning: No assignment found for row " + i);
            }
           
        }
         // Or, if the solver directly provides the min cost:
         // double minCost = solver.getOptimalCost(); 

        // 4. The maximum sum is the negative of the minimum cost
        return -minCost;
    }

    public static void main(String[] args) {
        // The 15x15 matrix from the image
        double[][] matrix = {
            {  7,  53, 183, 439, 863, 497, 383, 563,  79, 973, 287,  63, 343, 169, 583},
            {627, 343, 773, 959, 943, 767, 473, 103, 699, 303, 957, 703, 583, 639, 913},
            {447, 283, 463,  29,  23, 487, 463, 993, 119, 883, 327, 493, 423, 159, 743},
            {217, 623,   3, 399, 853, 407, 103, 983,  89, 463, 290, 516, 212, 462, 350},
            {960, 376, 682, 962, 300, 780, 486, 502, 912, 800, 250, 346, 172, 812, 350},
            {870, 456, 192, 162, 593, 473, 915,  45, 989, 873, 823, 965, 425, 329, 803},
            {973, 965, 905, 919, 133, 673, 665, 235, 509, 613, 673, 815, 165, 992, 326},
            {322, 148, 972, 962, 286, 255, 941, 541, 265, 323, 925, 281, 601,  95, 973},
            {445, 721,  11, 525, 473,  65, 511, 164, 138, 672,  18, 428, 154, 448, 848},
            {414, 456, 310, 312, 798, 104, 566, 520, 302, 248, 694, 976, 430, 392, 198},
            {184, 829, 373, 181, 631, 101, 969, 613, 840, 740, 778, 458, 284, 760, 390},
            {821, 461, 843, 513,  17, 901, 711, 993, 293, 157, 274,  94, 192, 156, 574},
            { 34, 124,   4, 878, 450, 476, 712, 914, 838, 669, 875, 299, 823, 329, 699},
            {815, 559, 813, 459, 522, 788, 168, 586, 966, 232, 308, 833, 251, 631, 107},
            {813, 883, 451, 509, 615,  77, 281, 613, 459, 205, 380, 274, 302,  35, 805}
        };

        try {
            double maxMatrixSum = calculateMaxMatrixSum(matrix);
            // Use long for the result if expecting large integer sums
            System.out.println("The Matrix Sum is: " + (long)maxMatrixSum); 
        } catch (Exception e) {
            System.err.println("Error calculating Matrix Sum: " + e.getMessage());
            e.printStackTrace();
        }
    }
}



class HungarianSolver {
    private double[][] costMatrix;
    private int rows, cols;
    private int[] assignment; // assignment[row] = assigned column
    private double minCost;
    private boolean[] rowCovered;
    private boolean[] colCovered;
    private double[][] originalCostMatrix;
    private int[][] marked; // 0 = unmarked, 1 = starred, 2 = primed

    // For numerical stability when comparing doubles
    private static final double EPSILON = 1e-10;
    private int originalRows, originalCols;

    public HungarianSolver(double[][] matrix) {
        // Store original dimensions
        this.originalRows = matrix.length;
        this.originalCols = matrix[0].length;

        // Clone original matrix
        this.originalCostMatrix = new double[originalRows][originalCols];
        for (int i = 0; i < originalRows; i++) {
            System.arraycopy(matrix[i], 0, originalCostMatrix[i], 0, originalCols);
        }

        // Make matrix square - use max dimension
        int maxDim = Math.max(originalRows, originalCols);
        this.rows = maxDim;
        this.cols = maxDim;

        this.costMatrix = new double[maxDim][maxDim];

        // Fill with large values (effectively infinity for the algorithm)
        for (int i = 0; i < maxDim; i++) {
            for (int j = 0; j < maxDim; j++) {
                costMatrix[i][j] = Double.MAX_VALUE / 2; // Avoid overflow in calculations
            }
        }

        // Copy actual values
        for (int i = 0; i < originalRows; i++) {
            for (int j = 0; j < originalCols; j++) {
                costMatrix[i][j] = matrix[i][j];
            }
        }

        this.assignment = new int[maxDim];
        this.rowCovered = new boolean[maxDim];
        this.colCovered = new boolean[maxDim];
        this.marked = new int[maxDim][maxDim];
    }

    public int[] execute() {
        // Initialize assignments
        Arrays.fill(assignment, -1);

        // Clear markings
        for (int i = 0; i < rows; i++) {
            Arrays.fill(marked[i], 0);
        }

        // Step 1: Reduce the matrix
        reduceMatrix();

        // Step 2: Initial star zeros
        initialStarZeros();

        // Main loop
        boolean done = false;
        while (!done) {
            // Step 3: Cover starred columns
            coverColumns();

            // Check if we're done
            int coveredColumns = countCoveredColumns();
            if (coveredColumns >= Math.min(rows, cols)) {
                done = true;
                continue;
            }

            // Step 4 & 5: Prime zeros and augment path
            boolean pathFound = findAugmentingPath();

            if (pathFound) {
                // Step 5: Augment path was successful, go back to step 3
                clearCovers();
                clearPrimes();
            } else {
                // Step 6: Adjust matrix
                adjustMatrix();
            }
        }

        // Extract the assignment from the starred zeros
        int[] result = new int[originalRows];
        Arrays.fill(result, -1);

        for (int i = 0; i < originalRows; i++) {
            for (int j = 0; j < originalCols; j++) {
                if (marked[i][j] == 1) {
                    result[i] = j;
                    break;
                }
            }
        }

        // Calculate optimal cost
        calculateMinCost(result);

        return result;
    }

    private void reduceMatrix() {
        // Subtract row minimums
        for (int i = 0; i < rows; i++) {
            double minValue = Double.MAX_VALUE;
            for (int j = 0; j < cols; j++) {
                if (costMatrix[i][j] < minValue) {
                    minValue = costMatrix[i][j];
                }
            }

            // Only subtract if minimum is finite
            if (minValue < Double.MAX_VALUE / 2) {
                for (int j = 0; j < cols; j++) {
                    if (costMatrix[i][j] < Double.MAX_VALUE / 2) {
                        costMatrix[i][j] -= minValue;
                    }
                }
            }
        }

        // Subtract column minimums
        for (int j = 0; j < cols; j++) {
            double minValue = Double.MAX_VALUE;
            for (int i = 0; i < rows; i++) {
                if (costMatrix[i][j] < minValue) {
                    minValue = costMatrix[i][j];
                }
            }

            // Only subtract if minimum is finite
            if (minValue < Double.MAX_VALUE / 2) {
                for (int i = 0; i < rows; i++) {
                    if (costMatrix[i][j] < Double.MAX_VALUE / 2) {
                        costMatrix[i][j] -= minValue;
                    }
                }
            }
        }
    }

    private void initialStarZeros() {
        // Track if a row or column already has a star
        boolean[] rowStarred = new boolean[rows];
        boolean[] colStarred = new boolean[cols];

        // First pass: try to star as many zeros as possible
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < cols; j++) {
                if (isZero(costMatrix[i][j]) && !rowStarred[i] && !colStarred[j]) {
                    marked[i][j] = 1; // Star this zero
                    rowStarred[i] = true;
                    colStarred[j] = true;
                }
            }
        }
    }

    private void coverColumns() {
        // Reset all covers
        Arrays.fill(rowCovered, false);
        Arrays.fill(colCovered, false);

        // Cover all columns with starred zeros
        for (int j = 0; j < cols; j++) {
            for (int i = 0; i < rows; i++) {
                if (marked[i][j] == 1) {
                    colCovered[j] = true;
                    break;
                }
            }
        }
    }

    private int countCoveredColumns() {
        int count = 0;
        for (boolean covered : colCovered) {
            if (covered) count++;
        }
        return count;
    }

    private boolean findAugmentingPath() {
        int row = -1, col = -1;

        // Find an uncovered zero
        int[] zeroPos = findUncoveredZero();
        row = zeroPos[0];
        col = zeroPos[1];

        if (row == -1) {
            // No uncovered zeros left
            return false;
        }

        // Mark the uncovered zero as primed
        marked[row][col] = 2;

        // Find a starred zero in the row
        int starCol = findStarInRow(row);

        if (starCol == -1) {
            // No starred zero in this row
            // Start the augmenting path with this zero
            int[] pathStart = {row, col};
            List<int[]> path = new ArrayList<>();
            path.add(pathStart);

            // Augment the matching along this path
            augmentPath(path);
            return true;
        } else {
            // There is a starred zero in this row
            // Cover this row and uncover the column with the starred zero
            rowCovered[row] = true;
            colCovered[starCol] = false;

            // Continue looking for augmenting path
            return continueFindingPath();
        }
    }

    private boolean continueFindingPath() {
        List<int[]> path = new ArrayList<>();
        int pathRow = -1, pathCol = -1;
        boolean done = false;

        // Start with the last primed zero that doesn't have a starred zero in its row
        for (int j = 0; j < cols; j++) {
            if (!colCovered[j]) {
                for (int i = 0; i < rows; i++) {
                    if (marked[i][j] == 2) {
                        // Found a primed zero in uncovered column
                        int starCol = findStarInRow(i);
                        if (starCol == -1) {
                            // No starred zero in this row - this is our starting point
                            pathRow = i;
                            pathCol = j;
                            done = true;
                            break;
                        }
                    }
                }
            }
            if (done) break;
        }

        if (pathRow == -1) {
            // No primed zeros with no starred zero in their row
            // Try to find any uncovered zero
            int[] zeroPos = findUncoveredZero();
            int row = zeroPos[0];
            int col = zeroPos[1];

            if (row == -1) {
                // No uncovered zeros left
                return false;
            }

            // Mark this zero as primed
            marked[row][col] = 2;

            // Find starred zero in this row
            int starCol = findStarInRow(row);

            if (starCol == -1) {
                // No starred zero - start augmenting path here
                pathRow = row;
                pathCol = col;
            } else {
                // Cover this row and uncover the column with the starred zero
                rowCovered[row] = true;
                colCovered[starCol] = false;
                return continueFindingPath();
            }
        }

        if (pathRow != -1) {
            // Found a starting point for augmenting path
            path.add(new int[]{pathRow, pathCol});
            augmentPath(path);
            return true;
        }

        return false;
    }

    private void augmentPath(List<int[]> initialPath) {
        // The path starts with a primed zero
        List<int[]> completePath = buildCompletePath(initialPath.get(0));

        // Augment the path
        for (int[] pos : completePath) {
            int r = pos[0];
            int c = pos[1];

            if (marked[r][c] == 1) {
                // Unstar starred zeros
                marked[r][c] = 0;
            } else if (marked[r][c] == 2) {
                // Star primed zeros
                marked[r][c] = 1;
            }
        }

        // Clear all covers and primes
        clearCovers();
        clearPrimes();
    }

    private List<int[]> buildCompletePath(int[] start) {
        List<int[]> path = new ArrayList<>();
        path.add(start);

        int row = start[0];
        int col = start[1];

        while (true) {
            // Find a starred zero in the column
            int starRow = findStarInColumn(col);

            if (starRow == -1) {
                // No starred zero in this column - path is complete
                break;
            }

            // Add starred zero to path
            path.add(new int[]{starRow, col});

            // Find a primed zero in the row
            col = findPrimeInRow(starRow);

            // Add primed zero to path
            path.add(new int[]{starRow, col});
        }

        return path;
    }

    private int findStarInRow(int row) {
        for (int j = 0; j < cols; j++) {
            if (marked[row][j] == 1) {
                return j;
            }
        }
        return -1;
    }

    private int findStarInColumn(int col) {
        for (int i = 0; i < rows; i++) {
            if (marked[i][col] == 1) {
                return i;
            }
        }
        return -1;
    }

    private int findPrimeInRow(int row) {
        for (int j = 0; j < cols; j++) {
            if (marked[row][j] == 2) {
                return j;
            }
        }
        return -1; // Should not happen in a valid path
    }

    private int[] findUncoveredZero() {
        for (int i = 0; i < rows; i++) {
            if (!rowCovered[i]) {
                for (int j = 0; j < cols; j++) {
                    if (!colCovered[j] && isZero(costMatrix[i][j])) {
                        return new int[]{i, j};
                    }
                }
            }
        }
        return new int[]{-1, -1}; // No uncovered zero found
    }

    private void clearCovers() {
        Arrays.fill(rowCovered, false);
        Arrays.fill(colCovered, false);
    }

    private void clearPrimes() {
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < cols; j++) {
                if (marked[i][j] == 2) {
                    marked[i][j] = 0;
                }
            }
        }
    }

    private void adjustMatrix() {
        // Find minimum uncovered value
        double minValue = findMinUncovered();

        // Add minValue to every covered row
        for (int i = 0; i < rows; i++) {
            if (rowCovered[i]) {
                for (int j = 0; j < cols; j++) {
                    if (costMatrix[i][j] < Double.MAX_VALUE / 2) {
                        costMatrix[i][j] += minValue;
                    }
                }
            }
        }

        // Subtract minValue from every uncovered column
        for (int j = 0; j < cols; j++) {
            if (!colCovered[j]) {
                for (int i = 0; i < rows; i++) {
                    if (costMatrix[i][j] < Double.MAX_VALUE / 2) {
                        costMatrix[i][j] -= minValue;
                    }
                }
            }
        }
    }

    private double findMinUncovered() {
        double min = Double.MAX_VALUE;

        for (int i = 0; i < rows; i++) {
            if (!rowCovered[i]) {
                for (int j = 0; j < cols; j++) {
                    if (!colCovered[j] && costMatrix[i][j] < min) {
                        min = costMatrix[i][j];
                    }
                }
            }
        }

        return min;
    }

    private boolean isZero(double value) {
        return Math.abs(value) < EPSILON;
    }

    public double getOptimalCost() {
        return minCost;
    }

    private void calculateMinCost(int[] result) {
        minCost = 0;

        for (int i = 0; i < originalRows; i++) {
            int j = result[i];
            if (j != -1 && j < originalCols) {
                minCost += originalCostMatrix[i][j];
            }
        }
    }
}

