import java.util.Arrays;
import java.text.DecimalFormat;

/**
 * This program calculates the expected number of rounds until a player gets both dice
 * in a circular dice-passing game with an even number of players.
 * 
 * Game rules:
 * - Players sit around a table in a circle
 * - Game begins with 2 dice held by opposite players
 * - On each turn, the two players with dice roll them
 * - If a player rolls 1, they pass the die to their left neighbor
 * - If a player rolls 6, they pass the die to their right neighbor
 * - Otherwise (2-5), they keep the die for the next turn
 * - Game ends when one player has both dice after rolling and passing
 * - The player with both dice loses
 */
public class DiceGameMarkovAnalysis {

    /**
     * Main method to run the program
     */
    public static void main(String[] args) {
        // // // Example usage with 4 players
        // int numPlayers = 4;
        // System.out.println("Calculating expected rounds for " + numPlayers + " players...");
        // double expectedRounds = solveExpectedRounds(numPlayers, 0, numPlayers/2);
        // System.out.println("Expected number of rounds: " + expectedRounds);
        
        // Uncomment to test with more players (note: performance degrades rapidly with larger player counts)
        long startTime, endTime;
        double result;
        int[] playerCounts = {10, 20, 30, 40, 50};
        for (int n : playerCounts) {
            System.out.println("\nCalculating for " + n + " players:");
            startTime = System.currentTimeMillis();
            result = solveExpectedRounds(n, 0, n/2);
            endTime = System.currentTimeMillis();
            DecimalFormat df = new DecimalFormat("#.##########"); // Aim for 10 decimal places
            System.out.println("Expected Number of rounds: " + df.format(result));
            System.out.println("Execution time: " + (endTime - startTime) / 1000.0 + " seconds");
        }
    }

    /**
     * Solves for the expected number of rounds until the game ends
     * 
     * @param n The number of players (must be even)
     * @param startI The position of the first die
     * @param startJ The position of the second die
     * @return The expected number of rounds until the game ends
     */
    public static double solveExpectedRounds(int n, int startI, int startJ) {
        // Validate parameters
        if (n % 2 != 0) {
            throw new IllegalArgumentException("Number of players must be even");
        }
        
        if (startI == startJ) {
            return 0; // Game already over
        }
        
        // Build the Q matrix (transitions between transient states)
        double[][] Q = buildQMatrix(n);
        
        // Compute the fundamental matrix N = (I-Q)^(-1)
        double[][] N = computeFundamentalMatrix(Q);
        
        // Calculate expected steps to absorption
        double[] expected = multiplyMatrixByOnes(N);
        
        // Get the result for our starting state
        int startIndex = stateToIndex(startI, startJ, n);
        return expected[startIndex];
    }
    
    /**
     * Maps a state (i,j) to a unique index in our matrices
     * 
     * @param i Position of first die
     * @param j Position of second die
     * @param n Number of players
     * @return The index corresponding to state (i,j)
     */
    public static int stateToIndex(int i, int j, int n) {
        if (i == j) {
            return -1; // Absorbing state, not included in our matrices
        }
        
        // More efficient mapping for large n
        if (i < j) {
            return i * (n - 1) + (j - (i < j ? 1 : 0));
        } else {
            return i * (n - 1) + j;
        }
    }
    
    /**
     * Returns the state (i,j) corresponding to a given index
     * 
     * @param index The index in our matrix
     * @param n Number of players
     * @return The state (i,j) as a 2-element array
     */
    public static int[] indexToState(int index, int n) {
        int statesPerPlayer = n - 1;
        int i = index / statesPerPlayer;
        int j = index % statesPerPlayer;
        
        // Adjust j if needed to account for the missing self-state
        if (j >= i) {
            j += 1;
        }
        
        return new int[] {i, j};
    }
    
    /**
     * Builds the transition matrix Q between transient states
     * 
     * @param n Number of players
     * @return The Q matrix
     */
    public static double[][] buildQMatrix(int n) {
        int numTransientStates = n * (n - 1);
        double[][] Q = new double[numTransientStates][numTransientStates];
        
        // Initialize Q with zeros
        for (double[] row : Q) {
            Arrays.fill(row, 0.0);
        }
        
        // For each transient state
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                if (i != j) { // Ensure it's a transient state
                    int sourceIndex = stateToIndex(i, j, n);
                    
                    // Calculate transitions based on dice rolls
                    // For player i's die
                    double i_keeps = 4.0/6.0;  // Probability of rolling 2,3,4,5
                    double i_left = 1.0/6.0;   // Probability of rolling 1
                    double i_right = 1.0/6.0;  // Probability of rolling 6
                    
                    int i_left_pos = (i - 1 + n) % n;   // Left neighbor (circular)
                    int i_right_pos = (i + 1) % n;      // Right neighbor (circular)
                    
                    // For player j's die
                    double j_keeps = 4.0/6.0;
                    double j_left = 1.0/6.0;
                    double j_right = 1.0/6.0;
                    
                    int j_left_pos = (j - 1 + n) % n;
                    int j_right_pos = (j + 1) % n;
                    
                    // Calculate all possible transitions and their probabilities
                    
                    // Both dice stay in place
                    if (i != j) {  // Still transient (always true here)
                        int destIndex = stateToIndex(i, j, n);
                        Q[sourceIndex][destIndex] += i_keeps * j_keeps;
                    }
                    
                    // i's die stays, j's moves left
                    if (i != j_left_pos) {  // Check if not absorbing
                        int destIndex = stateToIndex(i, j_left_pos, n);
                        Q[sourceIndex][destIndex] += i_keeps * j_left;
                    }
                    
                    // i's die stays, j's moves right
                    if (i != j_right_pos) {
                        int destIndex = stateToIndex(i, j_right_pos, n);
                        Q[sourceIndex][destIndex] += i_keeps * j_right;
                    }
                    
                    // i's die moves left, j's stays
                    if (i_left_pos != j) {
                        int destIndex = stateToIndex(i_left_pos, j, n);
                        Q[sourceIndex][destIndex] += i_left * j_keeps;
                    }
                    
                    // i's die moves right, j's stays
                    if (i_right_pos != j) {
                        int destIndex = stateToIndex(i_right_pos, j, n);
                        Q[sourceIndex][destIndex] += i_right * j_keeps;
                    }
                    
                    // Both dice move
                    // i left, j left
                    if (i_left_pos != j_left_pos) {
                        int destIndex = stateToIndex(i_left_pos, j_left_pos, n);
                        Q[sourceIndex][destIndex] += i_left * j_left;
                    }
                    
                    // i left, j right
                    if (i_left_pos != j_right_pos) {
                        int destIndex = stateToIndex(i_left_pos, j_right_pos, n);
                        Q[sourceIndex][destIndex] += i_left * j_right;
                    }
                    
                    // i right, j left
                    if (i_right_pos != j_left_pos) {
                        int destIndex = stateToIndex(i_right_pos, j_left_pos, n);
                        Q[sourceIndex][destIndex] += i_right * j_left;
                    }
                    
                    // i right, j right
                    if (i_right_pos != j_right_pos) {
                        int destIndex = stateToIndex(i_right_pos, j_right_pos, n);
                        Q[sourceIndex][destIndex] += i_right * j_right;
                    }
                }
            }
        }
        
        return Q;
    }
    
    /**
     * Computes the fundamental matrix N = (I - Q)^(-1)
     * 
     * @param Q The transition matrix
     * @return The fundamental matrix N
     */
    public static double[][] computeFundamentalMatrix(double[][] Q) {
        int size = Q.length;
        double[][] I = new double[size][size];
        double[][] diff = new double[size][size];
        
        // Create identity matrix
        for (int i = 0; i < size; i++) {
            I[i][i] = 1.0;
        }
        
        // Calculate I - Q
        for (int i = 0; i < size; i++) {
            for (int j = 0; j < size; j++) {
                diff[i][j] = I[i][j] - Q[i][j];
            }
        }
        
        // Invert the matrix (I - Q)
        return invertMatrix(diff);
    }
    
    /**
     * Inverts a matrix using Gaussian elimination
     * 
     * @param matrix The matrix to invert
     * @return The inverted matrix
     */
    public static double[][] invertMatrix(double[][] matrix) {
        int n = matrix.length;
        double[][] augmented = new double[n][2 * n];
        
        // Create augmented matrix [A|I]
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                augmented[i][j] = matrix[i][j];
                augmented[i][j + n] = (i == j) ? 1.0 : 0.0;
            }
        }
        
        // Perform Gaussian elimination
        for (int i = 0; i < n; i++) {
            // Find pivot
            int maxRow = i;
            for (int k = i + 1; k < n; k++) {
                if (Math.abs(augmented[k][i]) > Math.abs(augmented[maxRow][i])) {
                    maxRow = k;
                }
            }
            
            // Swap rows
            if (maxRow != i) {
                for (int k = 0; k < 2 * n; k++) {
                    double temp = augmented[i][k];
                    augmented[i][k] = augmented[maxRow][k];
                    augmented[maxRow][k] = temp;
                }
            }
            
            // Scale pivot row
            double pivot = augmented[i][i];
            if (Math.abs(pivot) < 1e-10) {
                throw new ArithmeticException("Matrix is singular or nearly singular");
            }
            
            for (int j = 0; j < 2 * n; j++) {
                augmented[i][j] /= pivot;
            }
            
            // Eliminate other rows
            for (int k = 0; k < n; k++) {
                if (k != i) {
                    double factor = augmented[k][i];
                    for (int j = 0; j < 2 * n; j++) {
                        augmented[k][j] -= factor * augmented[i][j];
                    }
                }
            }
        }
        
        // Extract the inverse matrix from the augmented matrix
        double[][] inverse = new double[n][n];
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                inverse[i][j] = augmented[i][j + n];
            }
        }
        
        return inverse;
    }
    
    /**
     * Multiplies a matrix by a column vector of ones
     * 
     * @param matrix The matrix to multiply
     * @return The resulting column vector
     */
    public static double[] multiplyMatrixByOnes(double[][] matrix) {
        int rows = matrix.length;
        int cols = matrix[0].length;
        double[] result = new double[rows];
        
        for (int i = 0; i < rows; i++) {
            double sum = 0.0;
            for (int j = 0; j < cols; j++) {
                sum += matrix[i][j];
            }
            result[i] = sum;
        }
        
        return result;
    }
    
    /**
     * For large numbers of players, this method uses a more efficient
     * approach that exploits symmetry and sparse representation
     * 
     * @param n The number of players
     * @param startI The position of the first die
     * @param startJ The position of the second die
     * @return The expected number of rounds until the game ends
     */
    public static double solveExpectedRoundsLarge(int n, int startI, int startJ) {
        // For demonstration purposes, this method is just a placeholder
        // that calls the standard method. In a real implementation, this would
        // use sparse matrices and more efficient algorithms.
        return solveExpectedRounds(n, startI, startJ);
        
        // Implementation for large n would use:
        // 1. Sparse matrix representation
        // 2. Iterative solvers for linear systems
        // 3. Symmetry exploitation
    }
    
    /**
     * Prints a matrix for debugging purposes
     * 
     * @param matrix The matrix to print
     */
    public static void printMatrix(double[][] matrix) {
        for (double[] row : matrix) {
            for (double value : row) {
                System.out.printf("%.4f ", value);
            }
            System.out.println();
        }
    }
}
