import collections
import time

def simulate_bean_game(initial_bowls):
    """
    Simulates the bean game efficiently using a dictionary for state
    and a set for eligible move tracking.

    Args:
        initial_bowls: A dictionary {index: count} representing the initial
                       number of beans in each bowl. Bowls not in the dict
                       are assumed to be empty.

    Returns:
        A dictionary {index: count} representing the final state where
        all counts are 1. Keys with 0 count are excluded.
        Returns it sorted by index for consistency.
    """
    # Use defaultdict for easier handling of empty bowls
    state = collections.defaultdict(int)
    state.update(initial_bowls)

    # Keep track of bowls that can make a move (count >= 2)
    # A set provides efficient add, remove, and arbitrary element retrieval (pop)
    eligible_indices = set()
    for i, count in initial_bowls.items():
        if count >= 2:
            eligible_indices.add(i)

    # --- Simulation loop ---
    while eligible_indices:
        # Pick an index to process. Using pop() from set is arbitrary but fine.
        # The uniqueness of the final state ensures the order doesn't matter.
        i = eligible_indices.pop()

        # Since we manage the set carefully, we can assume state[i] >= 2 here.
        # If state[i] somehow dropped below 2 without i being removed from
        # eligible_indices, that would indicate a bug in the update logic.

        # --- Perform the move ---
        # Store previous counts of neighbors *before* incrementing
        prev_neighbor_counts = {
            i - 1: state[i - 1], # Using defaultdict simplifies access
            i + 1: state[i + 1]
        }

        state[i] -= 2
        state[i - 1] += 1
        state[i + 1] += 1
        # --- Move performed ---


        # --- Update eligibility ---

        # Check bowl i (the one where the move happened)
        if state[i] >= 2:
            eligible_indices.add(i) # Still eligible, add it back

        # Check bowl i-1 (left neighbor)
        idx_minus_1 = i - 1
        # If the count was < 2 before and is now >= 2, it becomes eligible
        if prev_neighbor_counts[idx_minus_1] < 2 and state[idx_minus_1] >= 2:
            eligible_indices.add(idx_minus_1)

        # Check bowl i+1 (right neighbor)
        idx_plus_1 = i + 1
        # If the count was < 2 before and is now >= 2, it becomes eligible
        if prev_neighbor_counts[idx_plus_1] < 2 and state[idx_plus_1] >= 2:
            eligible_indices.add(idx_plus_1)

        # Note: We don't explicitly remove bowls when their count drops below 2 here.
        # They are implicitly removed by not being added back to eligible_indices
        # after their turn (via pop) if their count drops below 2.

    # --- Simulation complete ---

    # Clean up the final state: remove zero-count bowls explicitly
    # (although defaultdict wouldn't store them unless explicitly set to 0)
    # Also ensures the result is a standard dict.
    final_state = {idx: count for idx, count in state.items() if count > 0}

    # Verify final state properties (all counts should be 1)
    for idx, count in final_state.items():
        if count != 1:
             # This indicates a problem if the game theory holds true.
             raise ValueError(f"Simulation ended unexpectedly: Bowl {idx} has count {count}")

    # Return sorted dictionary for consistent output
    return dict(sorted(final_state.items()))

# --- Helper Function for Invariants (for testing) ---
def calculate_invariants(bowls):
    """Calculates the total number of beans (L) and the moment (M)."""
    L = sum(bowls.values())
    M = sum(i * count for i, count in bowls.items())
    return L, M

# --- Run the simulation for the given initial state ---
def simulate(initial_state, debug=True):
    if debug:
        print("Initial State:", initial_state)
    start_time = time.time()
    final_state = simulate_bean_game(initial_state)
    end_time = time.time()
    if debug:
        print("Final State:", sorted([int(k) for k in final_state.keys() if final_state[k]==1]))
    print(f"(Took {end_time - start_time:.6f} seconds)")
    
    # Verify Invariants
    initial_L, initial_M = calculate_invariants(initial_state)
    final_L, final_M = calculate_invariants(final_state)
    if debug:
        print(f"Initial Invariants: L={initial_L}, M={initial_M}")
        print(f"Final Invariants:   L={final_L}, M={final_M}")
    assert initial_L == final_L and initial_M == final_M
    if debug:
        print("Invariants preserved: Yes")
    print()


print("--- Example from Problem ---")
# Bowls at index 2 and 3 with 2 and 3 beans respectively
initial_state_example = {2: 2, 3: 3}
simulate(initial_state_example)

print("\n--- Single Bowl Example ---")
initial_state_complex = {7: 100}
simulate(initial_state_complex)

print("\n--- Complex Example ---")
initial_state_complex = {0: 10, 80: 10}
simulate(initial_state_complex)

print("\n--- Spacing Example ---")
initial_state_complex = {0: 10, 1: 10}
simulate(initial_state_complex)

initial_state_complex = {0: 10, 2: 10}
simulate(initial_state_complex)

initial_state_complex = {0: 10, 3: 10}
simulate(initial_state_complex)

initial_state_complex = {0: 10, 4: 10}
simulate(initial_state_complex)

initial_state_complex = {0: 10, 5: 10}
simulate(initial_state_complex)

### print("\n--- Large Number Examples ---")
### # Put N beans in bowl 0. Expect final state to be spread out.
### 
### print("N = 10:")
### initial_state_large = {0: 10}
### simulate(initial_state_large, False)
### 
### print("N = 100:")
### initial_state_large = {0: 100}
### simulate(initial_state_large, False)
### 
### print("N = 1000:")
### initial_state_large = {0: 1000}
### simulate(initial_state_large, False)

