import time
import math

def calculate_invariants(bowls):
    """Calculates the total number of beans (L) and the moment (M)."""
    L = sum(bowls.values())
    M = sum(i * count for i, count in bowls.items())
    return L, M

def calculate_final_state_directly(initial_bowls):
    """
    Calculates the final bean game state directly using invariants L and M,
    based on minimizing the sum of squared indices.

    Args:
        initial_bowls: A dictionary {index: count} representing the initial state.

    Returns:
        A dictionary {index: 1} representing the final state, sorted by index.
    """
    # 1. Calculate Invariants
    L, M = calculate_invariants(initial_bowls)

    if L == 0:
        return {}

    # 2. Determine Base Moment (for interval 0 to L-1)
    # Sum of 0 + 1 + ... + (L-1)
    M_base = L * (L - 1) // 2

    # 3. Calculate Moment Distribution
    M_offset = M - M_base
    q = M_offset // L  # Base shift for all indices
    r = M_offset % L   # Number of indices getting an extra +1 shift

    # Handle potential negative remainder in Python's % for negative M_offset
    # Ensure 0 <= r < L
    if r < 0:
        r += L
        q -= 1 # Adjust quotient accordingly if remainder was negative

    # 4. Construct Final Indices
    final_indices = set()
    # Indices shifted by q
    for i in range(L - r):
        final_indices.add(i + q)
    # Indices shifted by q+1 (the top r elements of the base sequence)
    for i in range(L - r, L):
        final_indices.add(i + q + 1)

    # 5. Return Result dictionary
    final_state = {idx: 1 for idx in final_indices}

    # Return sorted dictionary for consistent output
    return dict(sorted(final_state.items()))

# --- Example Usage ---
print("--- Example from Problem ---")
initial_state_example = {2: 2, 3: 3}
print("Initial State:", initial_state_example)
L_ex, M_ex = calculate_invariants(initial_state_example)
print(f"Invariants: L={L_ex}, M={M_ex}")
start_time = time.time()
final_state_example = calculate_final_state_directly(initial_state_example)
end_time = time.time()
print("Calculated Final State:", final_state_example)
print(f"(Took {end_time - start_time:.6f} seconds)")
# Verify matches previous simulation result {0: 1, 1: 1, 3: 1, 4: 1, 5: 1} - Yes!

# --- Complex Example ---
print("\n--- Complex Example ---")
initial_state_complex = {0: 5, 5: 4} # L=9, M=0*5+5*4=20
print("Initial State:", initial_state_complex)
L_comp, M_comp = calculate_invariants(initial_state_complex)
print(f"Invariants: L={L_comp}, M={M_comp}")
start_time = time.time()
final_state_complex = calculate_final_state_directly(initial_state_complex)
end_time = time.time()
print("Calculated Final State:", final_state_complex)
print(f"(Took {end_time - start_time:.6f} seconds)")

# --- Large Number Example ---
print("\n--- Large Number Example ---")
initial_state_large = {0: 1000000} # L=1,000,000, M=0
print("Initial State:", initial_state_large)
L_large, M_large = calculate_invariants(initial_state_large)
print(f"Invariants: L={L_large}, M={M_large}")
start_time = time.time()
final_state_large = calculate_final_state_directly(initial_state_large)
end_time = time.time()
# Don't print the full state, just confirm calculation is fast
print(f"Calculated Final State (Size: {len(final_state_large)})")
# Optionally print first/last few elements
keys = list(final_state_large.keys())
if len(keys) > 10:
    print(f"Example keys: {keys[:5]} ... {keys[-5:]}")
else:
     print(f"Final State Keys: {keys}")
print(f"(Took {end_time - start_time:.6f} seconds)")
