import java.util.*;

/*

Project Euler Problem 5

2520 is the smallest number that can be divided by each of the numbers from 1 to 10 without any remainder.

What is the smallest positive number that is evenly divisible by all of the numbers from 1 to 20?


Note:
The naive approach is to simply say,
multiply all 20 integers together.
That gives you 20!, a very large number.

The next approach is to find the smallest
number that is evenly divisible by 
all the integers i = 2 through n,
and if it is, print it out.

Unfortunately, while this integer is much
smaller than 20!, it is still too big 
to find in a brute force for loop.

There must be a better way!

Indeed, this problem boils down to finding the 
least common multiple, and a simple examination
of 5! = 120 reveals the right approach:

1*2*3*4*5 = 120 only has one repeated prime factor, a 2.

The least common multiple also happens 
to be small enough to find in our head, 
namely, 60, or 120/2.

So, we find any duplicate primes, like 
2 and 2^2, and replace them with just 2^2.

If we find the prime representation of each integer,
then we find the minimum prime representation 
shared by all of these integers, we will find
our least common multiple.

This will require getting a list of prime numbers 
up to 20 (whew!), and the prime factors representation 
of a given integer (easy to do if we can get all the primes
up to that number).
*/

public class Problem005 { 

	/** Find the minimum prime representation of integers 1 through 20. */
	public static void main(String[] args){ 
		minPrimeRep(1,20);
	}


	/** This function utilizes the prime factorizations
	 * of numbers 1 to N, and finds the minimum 
	 * number of primes required to represent 
	 * each number.
	 *
	 * Solutions to this question enable us
	 * to find least common multiples.
	 */
	public static void minPrimeRep(int MIN, int MAX) {
		if(MIN<1) throw new IllegalArgumentException("Error: MIN < 2.");
		if(MAX<MIN) throw new IllegalArgumentException("Error: MAX < MIN.");

		Map<Integer,Integer> primeExp = new TreeMap<Integer,Integer>();
		for(int i=MIN; i<=MAX; i++) {

			// map format is { <the prime> : <its exponent> }
			Map<Integer,Integer> thisPrimeExp = primeFactorizationExponentMap(i);
			
			// For each prime number in this integer's prime representation,
			for(Integer p : thisPrimeExp.keySet()) { 
				// Use a power equal to the maximum of these two:
				// - our current power of this prime
				// - the power of this prime in this integer
				//
				// Current power of this prime:
				Integer currpower;
				if( primeExp.keySet().contains(p)) {
					currpower = primeExp.get(p);
				} else {
					currpower = 0;
				}
				// Power of this prime contained in this integer
				Integer thispower = thisPrimeExp.get(p);
				
				Integer result = Math.max( currpower, thispower );
				primeExp.put(p, result);
			}
		}

		long ceiling = factorial((long)(MAX)); 
		System.out.println("The naive way to find a number divisible by integers 1..20");
		System.out.println("is to find 1*2*3*...*20 = 20!");
		System.out.println();
		System.out.println("20! = "+ceiling);
		System.out.println();
		System.out.println("Instead, we computed the prime representation of all the numbers 1-20,");
		System.out.println("and found the minimum subset of primes needed to represent all 20 integers:");
		System.out.println();
		Integer product = 1;
		for(Integer p : primeExp.keySet()) {
			Integer temp = p;
			for(Integer i=1; i<primeExp.get(p); i+=1) { 
				temp *= p;
			}
			product *= temp;
		}
		System.out.println("LCM = "+product);
		System.out.println();
	}


	/** Utilize the prime factorization method, defined below,
	 * which returns a simple list of prime factors,
	 * to build a map of each prime number and its corresponding exponent.
	 *
	 * This is a utility method for building the minimum 
	 * prime representation, and is just a wrapper method 
	 * for the prime factorization method below.
	 */
	public static Map<Integer,Integer> primeFactorizationExponentMap(int c) { 

		// Get the prime factorization as a list
		List<Integer> primes = primeFactorization(c);

		// Now turn it into a map:
		// { <the prime> : <its exponent> }
		// 2, 2, 2 becomes {2 : 3 }
		Map<Integer,Integer> exmap = new TreeMap<Integer,Integer>();

		for(Integer p : primes) { 
			if(exmap.keySet().contains(p)) { 
				Integer exponent = exmap.get(p);
				exponent += 1;
				exmap.put(p, exponent);

			} else {
				exmap.put(p,new Integer(1));

			}
		}

		return exmap;
	}


	/** This function computes the prime factorization
	 * of a number c.
	 */
	public static List<Integer> primeFactorization(int c) {
		if(c<2) { 
			List<Integer> primes = new ArrayList<Integer>();
			primes.add(c);
			return primes;
		}

		// List of primes up to c
		List<Integer> primes = Sieve.primeSieve(c);

		// arraylist - but errywhere else it don't matter what kinda list it is
		List<Integer> primeFactors = new ArrayList<Integer>();
		
		Integer a = new Integer(c);
		for (Integer b : primes) { 
			while(a != 1){
				if( a%b==0 ) {
					a /= b;
					primeFactors.add(b);
				} else {
					b++;
				}
			}
		}
		return primeFactors;
	}	



	/** Factorial function: n! = n*(n-1)*(n-2)*etc.
	 */
	public static long factorial(long n) { 
		if(n<0) { return 0; }
		else if(n==0||n==1) { return 1;}
		else {
			return n*factorial(n-1);
		}
	}

}



