import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;

/** Prime Digits Replacements: Example.
 *
 * Find the smallest prime which, 
 * by replacing part of the number 
 * (not necessarily adjacent digits) 
 * with the same digit, is part of an
 * eight prime value family.
 *
 * Eight prime value family means:
 * when you perform the operation 
 * on the specified pattern,
 * the result is a pool of 8 prime numbers.
 *
 * NOTE:
 *
 * This program only solves the example case,
 * of *3, and does not do 2-digit replacement.
 */
public class PrimeDigitReplacementExample {

	public static Set<Long> primes;
	public static final int POOLSIZE = 6;
	
	/** Find the candidate number[s] with the given prime pool. */
	public static void main(String[] args) { 
		int M = 2; // size of the candidates, *X
		poolFinder(POOLSIZE,M);
	}

	/** Quick test to make sure the sieve is working ok. */
	public static void testSieve() { 
		primes = primeSieve(100);
		System.out.println(primes);
	}

	/** Find the candidate numbers XX*X*XX that have a given prime pool size. */
	public static void poolFinder(int targetPoolSize, int M) { 

		primes = primeSieve((int)Math.pow(10,M));

		char[] mystuff = new char[M];
		for(int i=0; i<mystuff.length-1; i++) {

			// Make the recursive call to the method 
			char[] result = explore( mystuff, i, 0, targetPoolSize );

			// If we found something, print it out
			if(result!=null) { 
				System.out.println("Found a number with a prime pool of "+targetPoolSize);
				System.out.println(Arrays.toString(result));
				System.out.print("\n");
				break;
			}
		}
	}

	/** Recursive method: fill in the candidate number XX**X with digits,
	 * and once it is full, compare the size of the prime pool to targetPoolSize. */
	public static char[] explore(char[] arr, int i, int depth, int targetPoolSize) { 
		//System.out.println("Calling explore:");
		//System.out.println("    arr = "+Arrays.toString(arr)); 
		//System.out.println("    i = "+i);
		//System.out.println("    depth = "+depth);
		int M = arr.length;
		if (depth>M) { 
			throw new IllegalArgumentException();
		} else if(depth==M) { 
			// Base case
			//System.out.println("    Base case...");
			arr[i] = '*';
			// the call below changes * to 9
			int pool = primePool(arr,i);
			//System.out.println("The pool for "+Arrays.toString(arr)+" is "+pool);
			if(pool==targetPoolSize) return arr; else return null;

		} else if(depth==i) {
			// Skip this digit recursive case
			//System.out.println("    Recursive (skip) case..."); 
			return explore(arr,i,depth+1, targetPoolSize);
		} else {
			// Handle this digit recursive case
			//
			// Note: 
			// Could also use Character.forDigit(i) 
			// http://docs.oracle.com/javase/7/docs/api/java/lang/Character.html#forDigit(int,%20int)
			//
			//System.out.println("    Recursive case..."); 
			for(char c = '0'; c <= '9'; c++) { 
				arr[depth] = c;
				char[] result = explore(arr, i, depth+1, targetPoolSize);
				if(result!=null) { 
					return result;
				}
			}
			return null;
		}
	}

	/** Find all prime numbers below n.  */
	public static Set<Long> primeSieve(int n) { 

        // initially assume all integers are prime 
        boolean[] isPrime = new boolean[n+1];
        for (int i = 2; i <= n; i++) {
            isPrime[i] = true;
        }

        // mark non-primes <= n using Sieve of Eratosthenes
        for (int factor = 2; factor*factor <= n; factor++) {

            // if factor is prime, then mark multiples of factor as nonprime
            // suffices to consider mutiples factor, factor+1, ...,  n/factor
            if (isPrime[factor]) {
                for (int j = factor; factor*j <= n; j++) {
                    isPrime[factor*j] = false;
                }
            }
        }

		// Add primes to list
		Set<Long> primes = new HashSet<Long>();
		int nprimes = 0;
		for(int k=2; k<=n; k++) {
			if(isPrime[k]) { 
				primes.add( new Long(k) );
				nprimes++;
			}
		}

		return primes;
	}

	/** For a number with a missing digit at index i, determine the size 
	 * of the prime number pool when the missing digit is filled with 0-9. */
	public static int primePool(char[] arr, int i) {
		// Start with a fresh copy
		char[] fresh = arr.clone();
		// Compute the prime pool
		int primePool = 0;
		// Note that a leading 0 doesn't count, as per the *3 example
		char startchar;
		if(i==0) startchar = '1'; else startchar = '0';

		// Replace i with 0/1 through 9, and check if each is a prime
		for(char c = startchar; c<='9'; c++){
			fresh[i] = c;
			Long candidate = Long.parseLong(new String(fresh));
			if(primes.contains(candidate)) { 
				primePool++;
			}
		}
		return primePool;
	}

}

