import sympy
import itertools
from typing import List, Tuple, Set

MAX_SMALL = 10000
MAX_FIFTH = 10000
MAX_PRIME = 10000000

def is_prime(n: int) -> bool:
    """Check if a number is prime using sympy's primality test."""
    return sympy.isprime(n)

def concatenate(a: int, b: int) -> int:
    """Concatenate number a with number b (as a string)."""
    return int(str(a) + str(b))

def check_prime_pair_property(primes: List[int]) -> bool:
    """
    Check if a list of primes satisfies the property that 
    any two primes concatenated (in either order) form a prime.
    """
    for p1, p2 in itertools.combinations(primes, 2):
        if not is_prime(concatenate(p1, p2)) or not is_prime(concatenate(p2, p1)):
            return False
    return True

def find_compatible_prime(existing_set: List[int], candidate_primes: List[int]) -> List[int]:
    """Find primes that are compatible with the existing set."""
    compatible_primes = []
    
    for prime in candidate_primes:
        if prime in existing_set:
            continue
            
        compatible = True
        for existing_prime in existing_set:
            if not is_prime(concatenate(prime, existing_prime)) or not is_prime(concatenate(existing_prime, prime)):
                compatible = False
                break
                
        if compatible:
            compatible_primes.append(prime)
            
    return compatible_primes

def find_lowest_sum_set(size: int = 5, max_prime: int = MAX_PRIME) -> Tuple[List[int], int]:
    """
    Find the set of 'size' primes with the lowest sum where
    any two primes concatenated form a prime.
    """
    # Generate prime numbers up to max_prime
    primes = list(sympy.primerange(2, max_prime))
    
    # Start with the known set of four primes
    known_set = [3, 7, 109, 673]
    
    if size == 4:
        return known_set, sum(known_set)
    
    # Find compatible primes to add to the known set
    compatible_primes = find_compatible_prime(known_set, primes)
    compatible_primes.sort()  # Sort by value to prioritize smaller primes
    
    if compatible_primes:
        best_prime = compatible_primes[0]
        best_set = known_set + [best_prime]
        return best_set, sum(best_set)
    
    # If no compatible prime found, try a different approach:
    # Look for sets of 5 primes from scratch
    best_sum = float('inf')
    best_set = []
    
    # Start with small primes for better performance
    small_primes = list(sympy.primerange(2, MAX_SMALL))
    
    # Try to build sets directly with the smallest primes
    for p1 in small_primes:  # Limit the search to small primes
        set1 = [p1]
        compatible1 = find_compatible_prime(set1, small_primes)
        
        for p2 in compatible1:
            if p2 <= p1:
                continue  # Skip to avoid duplicates
                
            set2 = [p1, p2]
            compatible2 = find_compatible_prime(set2, small_primes)
            
            for p3 in compatible2:
                if p3 <= p2:
                    continue
                    
                set3 = [p1, p2, p3]
                compatible3 = find_compatible_prime(set3, small_primes)
                
                for p4 in compatible3:
                    if p4 <= p3:
                        continue
                        
                    set4 = [p1, p2, p3, p4]
                    compatible4 = find_compatible_prime(set4, small_primes)
                    
                    for p5 in compatible4:
                        if p5 <= p4:
                            continue
                            
                        candidate_set = [p1, p2, p3, p4, p5]
                        candidate_sum = sum(candidate_set)
                        
                        if candidate_sum < best_sum:
                            # Verify the whole set
                            if check_prime_pair_property(candidate_set):
                                best_sum = candidate_sum
                                best_set = candidate_set
                                print(f"Found better set: {best_set} with sum {best_sum}")
    
    return best_set, best_sum

def search_for_compatible_fifth_prime(known_set: List[int], max_prime: int = MAX_FIFTH) -> Tuple[List[int], int]:
    """
    Search for a fifth prime that's compatible with the known set of four primes.
    Uses a more extensive search than the main function.
    """
    # Generate more prime numbers to test
    candidate_primes = list(sympy.primerange(2, max_prime))
    
    compatible_primes = find_compatible_prime(known_set, candidate_primes)
    compatible_primes.sort()  # Sort by value
    
    if compatible_primes:
        best_prime = compatible_primes[0]
        best_set = known_set + [best_prime]
        return best_set, sum(best_set)
    
    return [], 0

def main():
    print("Starting search for a set of five primes...")
    
    # First try with the known set of four primes
    known_set = [3, 7, 109, 673]
    result_set, sum_value = search_for_compatible_fifth_prime(known_set)
    
    if result_set:
        print(f"Found solution: {result_set}")
        print(f"Sum: {sum_value}")
    else:
        print("No solution found extending the known set. Trying from scratch...")
        
        # Try a more general search
        result_set, sum_value = find_lowest_sum_set(size=5)
        
        if result_set:
            print(f"Found solution: {result_set}")
            print(f"Sum: {sum_value}")
            
            # Verify all concatenations are prime
            all_pairs = list(itertools.combinations(result_set, 2))
            for p1, p2 in all_pairs:
                c1 = concatenate(p1, p2)
                c2 = concatenate(p2, p1)
                print(f"{p1} + {p2} = {c1} (Prime: {is_prime(c1)})")
                print(f"{p2} + {p1} = {c2} (Prime: {is_prime(c2)})")
        else:
            print("No solution found. Try increasing the search range.")

if __name__ == "__main__":
    main()

