import java.util.*;

/**
 * Figurate numbers solution:
 *
 * Find the sum of ordered set of six cyclic 4-digit numbers for which each polygonal type:
 * triangle, square, pentagonal, hexagonal, heptagonal, octagonal, is represented
 * by a different number in the set.
 */
public class CorrectedFigurate {
    
    // Number of different shapes we're looking for
    public static final int NSHAPES = 3;
    public static final int SHAPES_START = 3;
    //public static final int SHAPES_END = 8;
    public static final int SHAPES_END = 5;

    // Limit: only use figurate numbers of to MAX_FIGURATE
    public static final int MAX_FIGURATE = 10000;
    
    public static final int FOUR_DIGITS = 1000;

    enum Shape {
        TRIANGLE(3), SQUARE(4), PENTAGONAL(5); //HEXAGONAL(6), HEPTAGONAL(7), OCTAGONAL(8);
        final int value;
        Shape(int value) { this.value = value; }
    }
    
    public static void main(String[] args) {
        solve();
    }

    public static void solve() {
        // Generate all 4-digit figurate numbers for each shape
        Map<Shape, List<Integer>> allNumbers = new EnumMap<>(Shape.class);

        for (Shape shape : Shape.values()) {
            List<Integer> numbers = getAllFigurateNumbers(shape);
            allNumbers.put(shape, numbers);
            System.out.println("Shape " + shape + " has " + numbers.size() + " 4-digit numbers");
        }
        
        // Create maps from last two digits to list of numbers that end with those digits
        Map<Integer, List<FigurateNumber>> lastTwoDigitsMap = new HashMap<>();
        
        // Populate the map
        for (Shape shape : Shape.values()) {
            List<Integer> shapeNumbers = allNumbers.get(shape);
            for (int num : shapeNumbers) {
                FigurateNumber fn = new FigurateNumber(num, shape);
                int lastTwoDigits = num % 100;
                lastTwoDigitsMap.computeIfAbsent(lastTwoDigits, k -> new ArrayList<>()).add(fn);
            }
        }
        
        // Try each number as the starting point for the cyclic set
        for (Shape startShape : Shape.values()) {
            List<Integer> shapeNumbers = allNumbers.get(startShape);
            for (int startNum : shapeNumbers) {
                Set<Shape> usedShapes = new HashSet<>();
                usedShapes.add(startShape);

                List<FigurateNumber> path = new ArrayList<>();
                path.add(new FigurateNumber(startNum, startShape));
                
                if (findCyclicSet(path, usedShapes, lastTwoDigitsMap, startNum)) {
                    // Solution found, calculate and print the sum
                    int sum = 0;
                    System.out.println("Found solution:");
                    for (FigurateNumber fn : path) {
                        System.out.println(fn);
                        sum += fn.value;
                    }
                    System.out.println("Sum: " + sum);
                    return;
                }
            }
        }

        System.out.println("No solution found");
    }
    
    /**
     * Recursive function to find a cyclic set of figurate numbers
     */
    private static boolean findCyclicSet(List<FigurateNumber> path, 
                                         Set<Shape> usedShapes, 
                                         Map<Integer, List<FigurateNumber>> lastTwoDigitsMap, 
                                         int startNum) {
        // If we've found all shapes
        if (path.size() == NSHAPES) {
            // Check if the last number connects back to the first
            int lastNum = path.get(path.size() - 1).value;
            return (lastNum % 100) == (startNum / 100);
        }
        
        // Get the last number in our current path
        FigurateNumber lastFn = path.get(path.size() - 1);
        int lastTwoDigits = lastFn.value % 100;
        
        // Find numbers that start with these last two digits
        List<FigurateNumber> candidates = lastTwoDigitsMap.getOrDefault(lastTwoDigits, Collections.emptyList());
        for (FigurateNumber candidate : candidates) {
            if (!usedShapes.contains(candidate.shape)) {
                // Add to path and mark shape as used
                path.add(candidate);
                usedShapes.add(candidate.shape);

                // Recursive call
                if (findCyclicSet(path, usedShapes, lastTwoDigitsMap, startNum)) {
                    return true;
                }

                // Backtrack
                path.remove(path.size()-1);
                usedShapes.remove(candidate.shape);
            }
        }

        return false;
    }
    
    /**
     * Returns all 4-digit figurate numbers for the given shape
     */
    public static List<Integer> getAllFigurateNumbers(Shape shape) {
        List<Integer> result = new ArrayList<>();
        int term = 1;
        int value;

        while ((value = calculateFigurateNumber(term, shape)) < MAX_FIGURATE) {
            if (value >= FOUR_DIGITS) {
                result.add(value);
            }
            term++;
        }

        return result;
    }

    /**
     * Calculates the nth s-gonal figurate number correctly according to the problem formulas
     */
    public static int calculateFigurateNumber(int n, Shape shape) {
        if (shape == null) {
            throw new IllegalArgumentException("Shape cannot be null");
        }
        switch (shape) {
            case TRIANGLE: return n * (n + 1) / 2;
            case SQUARE: return n * n;
            case PENTAGONAL: return n * (3 * n - 1) / 2;
            //case HEXAGONAL: return n * (2 * n - 1);
            //case HEPTAGONAL: return n * (5 * n - 3) / 2;
            //case OCTAGONAL: return n * (3 * n - 2);
            default: throw new IllegalArgumentException("Invalid shape: " + shape);
        }
    }

    /**
     * Class to store a figurate number along with its shape
     */
    static class FigurateNumber {
        int value;
        Shape shape;
        
        public FigurateNumber(int value, Shape shape) {
            this.value = value;
            this.shape = shape;
        }
        
        @Override
        public String toString() {
            return value + " (shape=" + shape + ")";
        }
    }
}
