import java.util.*;
import com.google.common.graph.*;

/** Figurate numbers using Guava:
 *
 * Find the sum of ordered set of six cyclic 
 * 4-digit numbers for which each polygonal type,
 * triangle, square, pentagonal, hexagonal,
 * heptagonal, octagonal, is represented
 * by a different number in the set.
 *
 * Approach:
 *
 * Set up a graph, except 
 * NODES are the two-digits that connect numbers,
 * and EDGES are the numbers themselves.
 * Then we seek a path that spans all six partitions.
 *
 *
 */
public class Figurate { 

	public static final int NSHAPES = 8-3+1;

	public static void main(String[] args) { 
		solve();
	}

	public static void solve() { 
		MutableNetwork<Node,Edge> graph = buildNetwork();
		System.out.println("Number of nodes: "+graph.nodes().size());
		System.out.println("Number of edges: "+graph.edges().size());

		findPartitionCycle(graph);
	}





	/********************************
	 * Build the graph
	 * ******************************/

	/** Construct the graph containing all the shape numbers, with edges. */
	public static MutableNetwork<Node,Edge> buildNetwork() { 

		MutableNetwork<Node,Edge> cycle = NetworkBuilder.directed().build();

		//////////////////////////////////
		// Construct nodes

		Set<Integer> keys = new HashSet<Integer>();
		Map<Integer,Node> fixes = new HashMap<Integer,Node>();

		// Loop over all shape groups
		for(int s=3; s<3+NSHAPES; s++) { 
			Set<Integer> shapeNumbers = getAll(s);
			for(Integer sh : shapeNumbers) {
				int prefix  = sh/100;
				int postfix = sh%100;

				// Add prefix node
				Node preNode = null;
				if(fixes.keySet().contains(prefix)) { 
					preNode = fixes.get(prefix);
				} else {
					preNode = new Node(prefix);
					fixes.put(prefix,preNode);
				}

				// Add postfix node to graph
				Node postNode = null;
				if(postfix>9) { 
					if(fixes.keySet().contains(postfix)) { 
						postNode = fixes.get(postfix);
					} else {
						postNode = new Node(postfix);
						fixes.put(postfix, postNode);
					}

					// Now link them
					Edge edgy = new Edge(sh,s);

					// Add everything....... IF everything is there
					cycle.addNode(postNode);
					cycle.addNode(preNode);
					try {
						cycle.addEdge(preNode, postNode, edgy);
					} catch(IllegalArgumentException exc) { 
						continue;
					}
				}
			}
		}

		return cycle;
	}


	/*******************************
	 * Find a cycle that visits all 6 partitions
	 * *****************************/

	public static void findPartitionCycle(MutableNetwork<Node,Edge> graph) {

		Stack<Integer> cycle = new Stack<Integer>();
		boolean[] shapes = new boolean[9];

		// Loop over all nodes and do a DFS looking for a complete circuit
		for(Node vertex : graph.nodes()) {
			// Keep track of:
			// - graph
			// - node
			// - which shapes completed
			// - stack of integers with solution
			if( dfs(graph, vertex, shapes, cycle, 0) ) {
				System.out.println(cycle);
			}
		}

	}

	public static boolean dfs(MutableNetwork<Node,Edge> graph,
					Node vertex,
					boolean[] shapes,
					Stack<Integer> cycle,
					int depth) { 

		if(depth==1) { 
			System.out.println("Just checking: "+vertex);
		}

		int shapesSum = 0;
		for(Boolean b : shapes) {
			if(b) shapesSum++;
		}
		if(shapesSum==6) {
			// Base case:
			// Shapes is all true
			// Almost finished
			for(Node w : graph.adjacentNodes(vertex)) { 
				if(w.first) {
					// Populate cycle with edge values...
					System.out.println("Found path.");
					return true;
				}
			}
			return false;
		} else {
			for(Node w : graph.adjacentNodes(vertex)) { 
				if(!w.visited) { 
					int whichShape = 0;
					Set<Edge> connections = null;
					connections = graph.edgesConnecting(vertex,w);
					if(connections.size()==1) { 
						whichShape = connections.iterator().next().shape;
					} 
					connections = graph.edgesConnecting(vertex,w);
					if(connections.size()==1) { 
						whichShape = connections.iterator().next().shape;
					}

					if( !shapes[whichShape] ) {

						w.visit();
						shapes[whichShape] = true;
						cycle.push(w.value);

						if(dfs(graph, w, shapes, cycle, depth+1)) {
							return true;
						}

						w.unvisit();
						shapes[whichShape] = false;
						cycle.pop();
					}
				}
			}
			return false;
		}
	}





	/*******************************
	 * Shape function and inverse
	 * (used to construct graph)
	 * *****************************/

	/** Compute the shape function for an s-sided shape. */
	public static int shapeFunction(int n, int s) { 
		int a = s - 2;
		int b = s - 4;
		return (n*(a*n-b))/2;
	}

	/** Compute the inverse shape function for an s-sided shape. */
	public static double inverseShapeFunction(int limit, int s) { 
		int a = s - 2;
		int b = s - 4;
		return (b + Math.sqrt(b*b + 8*a*limit))/(2.0*a);
	}

	/** Get all 4-digit numbers corresponding to an s-sided shape. */
	public static Set<Integer> getAll(int s) { 
		int floor = 1000;
		int ceil = 9999;
		int start = (int)(Math.ceil(inverseShapeFunction(floor,s)));
		int end = (int)(Math.floor(inverseShapeFunction(ceil,s)));
		Set<Integer> result = new HashSet<>();
		for(int k=start; k<=end; k++) { 
			int num = shapeFunction(k,s);
			// anything with second to last digit of 0 requires 3-digit numbers.
			if( (num/10)%10 != 0) {
				result.add(num);
			}
		}
		return result;
	}


}




/** Node class */
class Node {
	public int value;
	public boolean visited;
	public boolean first;
	public Node(int n){
		value = n;
		visited = false;
	}
	public String toString() { 
		StringBuffer sb = new StringBuffer();
		sb.append(value);
		return sb.toString();
	}
	public void first() { 
		first = true;
	}
	public void visit() { 
		visited = true;
	}
	public void unvisit() { 
		visited = false;
	}
}

class Edge {
	public int shape;
	public int value;
	public Edge(int value, int shape) {
		this.value = value;
		this.shape = shape;
	}
}
