import java.util.List;
import java.util.ArrayList;
import java.util.Queue;
import java.util.PriorityQueue;

/** Factor Generator.
 *
 * This generates the primes and powers of primes 
 * in sequential order.
 *
 * It would be nice if this generated them
 * dynamically, but alas, the logic 
 * for keeping track, switching prime sieve algorithms, 
 * etc., is all a bit too much for this problem.
 */
public class FactorGenerator {

	int MAX;
	List<Long> primes = new ArrayList<Long>();
	PriorityQueue<Integer> pq = new PriorityQueue<Integer>();

	public FactorGenerator(int MAX) {
		this.MAX = MAX;
		populateQueue();
	}

	/** Populate the priority queue with prime numbers
	 * and their various power-of-two powers.
	 */
	protected void populateQueue() { 
		// Calculate the log base 2 of MAX -
		// that gives you the maximum power of prime you will need
		// (number of queues you will need).
		//
		// Example: if our number is 31, log2(31) < 5, 
		// so the smallest prime representation with maximum exponent 
		// can't be larger than 2^5. So, we don't need a 5th power queue.
		//int log2 = binlog(MAX);
		int log2 = (int)(Math.log(MAX)/Math.log(2));

		populatePrimes();

		// Now we choose which primes to add.
		for(Long prime : primes) { 
			for(int j=1; j<=log2; j++) { 
				// Check if this prime, or any of its powers, 
				// are possible factors.
				int p = prime.intValue();
				if( j < Math.log(MAX)/Math.log(p) ) {
					pq.add( intpow(p,j) );
				} else {
					// don't calculate 16th powers of 10 digit numbers, plz
					break;
				}
			}
		}

	} 

	/** Populate primes list. */
	private void populatePrimes() { 
		primes = Sieve.primeSieve(MAX);
	}

	/** Boolean - is there a next factor in the priority queue? */
	public boolean hasNext() { 
		return !pq.isEmpty();
	}

	/** Return the next factor. */
	public Integer next() {
		return pq.remove();
	}

	/** Manual integer power method. */
	private static int intpow(int a, int b) { 
		int prod = 1;
		for(int j=0; j<b; j++) { 
			prod *= a;
		}
		return prod;
	}
}

