import numpy as np

"""
Dyanmic Program

Given a set of RRRRRRs starting at a given row/column, 
construct all variant castles that are valid.

Example:


            # # 
            # # 
            # # 
            # # 
# # # # # # # # 
# # # # # # # # 
# # # # # # # # 
# # # # # # # # 

U U U U R R R R R R U


6 Rs form 5 slots
Slots are at a height of 4
3 squares to move down for initial move
H = 8


If we have 5 slots, 
we can have up to 4 slots occupied by Ds,
and up to 4 slots occupied by Us
"""

def main():
    castle_to_dynamic("UURUURRRRRRUUUURDDDRDDDDD")

def castle_to_dynamic(castle):
    num_u = sum([1 if j is 'U' else 0 for j in castle])
    num_d = sum([1 if j is 'D' else 0 for j in castle])
    num_r = sum([1 if j is 'R' else 0 for j in castle])

    H = num_u
    if(num_u is not num_d):
        raise Exception("number of Us != number of Ds (they must be equal)")

    min_run_len = 3

    # Assemble tuples for front, middle, and back:
    # ( start_index, height_of_run, length_of_run )
    # 
    # These are assembled with some slightly baffling
    # and magical (but also tested) list comprehensions
    # (they're kind of addictive that way)

    # Locations of Us
    us = [i for i,j in enumerate(castle) if j is 'U']

    # Length of runs of Rs between the H Us
    # Tuple: (index of run, length of run)
    urs = [(us[i]+1, j-1) for i,j in enumerate(np.diff(us)) if j-1 >= min_run_len ]

    # Locations of Ds
    ds = [i for i,j in enumerate(castle) if j is 'D']

    # Length of runs of Rs between the H Ds
    # Tuple: (index of run, lenght of run)
    drs = [(ds[i]+1, j-1) for i,j in enumerate(np.diff(ds)) if j-1 >= min_run_len]

    # Midpoint start and end
    mid_start = us[-1]
    mid_end = ds[0]
    mrs = []
    if( (mid_end - mid_start - 1) > min_run_len ):
        mrs = [(mid_start+1, H, mid_end-mid_start-1)]

    print(castle)
    print()

    print("Runs of Rs in front slots:")
    print(urs)
    print()

    for ur in urs:

        #print("Calling dynamic(%d, %d, %d)"%( *ur, H ))

        start_index = ur[0]
        run_length = ur[1]
        h = sum([1 for p in range(start_index) if castle[p] is 'U'])

        # We only choose N-1 heights, the last one is determined for us
        # (has to meet up with the other side)
        which_slot = 0
        slot_length = run_length - 1
        slot_choices = [None]*(slot_length-1)

        castle_container = set()

        castle_container = dynamic( h, H,
                                    which_slot, 
                                    slot_choices,
                                    castle,
                                    start_index, run_length,
                                    castle_container)

        # Finished

    print()

    print("Runs of Rs in middle slots:")
    print(mrs)
    print()

    print("Runs of Rs in back slots:")
    print(drs)
    print()




def dynamic(h, H,
            which_slot,
            slot_choices,
            original_castle,
            start_index, run_length,
            castle_container):
    """
    We have run_length slots to fill.
    If slot #1, h choices (1 thru h).
    If slot #2+, H choices (1 thru H).
    Base case is, slots are maxed out.

    h                : row height of this run of RRRs
    H                : max height of castle
    which_slot       : index of the slot we're choosing a height for
    slot_choices     : what heights have been chosen for previous slots
    original_castle  : string containing original castle
    start_index      : index in string where run of RRRs starts
    run_length       : length of this run of RRRs
    castle_container : container for all the castle variations generated
    """
    slot_length = run_length - 1

    # which_slot starts at 0 and goes to (slot length - 1)
    if(which_slot == (slot_length-1) ):

        # Construct the left and right of the run of RRRs
        l_piece = list(original_castle[:start_index])
        r_piece = list(original_castle[start_index+run_length:])

        # Construct the variation on the run of RRRs
        m_piece = []
        prev_height = h
        for choice in slot_choices:
            m_piece.append("R")
            moves = choice - prev_height
            if(moves<0):
                m_piece.append("D"*abs(moves))
            elif(moves>0):
                m_piece.append("U"*moves)
            prev_height = choice

        # Last step
        moves = h - prev_height
        m_piece.append("R")
        if(moves<0):
            m_piece.append("D"*abs(moves))
        elif(moves>0):
            m_piece.append("U"*moves)
        m_piece.append("R")

        new_castle = "".join(l_piece + m_piece + r_piece)
        castle_container.add(new_castle)

        #print(original_castle)
        print(new_castle)
        #print()

        return castle_container

    else:

        # If which_slot is 0, or if all slot
        # choices up to this point have been h
        # (zero height change),
        # this is our first slot
        first_slot = True
        if(which_slot>0):
            for i in range(1,which_slot):
                if(slot_choices[i] != slot_choices[i-1]):
                    first_slot = False

        if(first_slot):

            # We're dealing with the first downward steps

            # Possible column heights:
            # 
            # Case A:
            # h-1 choices (height 1 through h-1)
            # 
            # Case B:
            # 1 choice (height h)

            # range will cover everything 
            # up to (not including) MAX
            MAX = h

        else:

            # The first downward steps have been taken,
            # so now we can take either up or down steps.

            # Possible column heights:
            # 
            # Case A:
            # H - 1 choices (height 1 through H, excluding 0 change)
            # 
            # Case B:
            # 1 choice (height of 0 change)

            MAX = H


        for j in range(1,MAX+1):

            # Make this choice of height
            slot_choices[which_slot] = j

            # Explore consequences
            castle_container = dynamic( h, H,
                                        which_slot+1,
                                        slot_choices,
                                        original_castle,
                                        start_index, run_length,
                                        castle_container)

            # Unmake choice of height
            slot_choices[which_slot] = None

        return castle_container



if __name__=="__main__":
    main()
