import multiprocessing as mp
import time
import sys
import math

def find_2025_numbers_in_range(start_sum, end_sum, n):
    """
    Find all 2025-numbers generated from sums in the range [start_sum, end_sum]
    where the resulting number has at most n digits.
    """
    found_numbers = []
    
    # Pre-compute powers of 10 for efficiency
    powers_of_10 = [10**i for i in range(n+1)]
    
    for sum_val in range(start_sum, end_sum + 1):
        square = sum_val * sum_val
        
        # Skip if square has more than n digits
        if square >= powers_of_10[n]:
            break
            
        square_str = str(square)
        
        # Check all possible splits of the square to see if it's a 2025-number
        for i in range(1, len(square_str)):
            a = int(square_str[:i])
            b = int(square_str[i:])
            zero_case = (str(a)+str(b) != square_str) or (a==0) or (b==0)
            
            # Check if a + b = sum_val
            if a + b == sum_val and not zero_case:
                found_numbers.append((square, a, b, sum_val))
                break
    
    return found_numbers

def calculate_T_parallel(n):
    """Calculate T(n) using parallel processing for performance."""
    # Maximum possible sum value that could produce a 2025-number with n digits
    max_sum_value = int(math.sqrt(10**n)) + 1
    
    # Determine number of processes to use
    cpu_count = mp.cpu_count()
    
    # Split the work into chunks based on CPU count
    chunk_size = max(1, max_sum_value // cpu_count)
    ranges = []
    
    for i in range(0, max_sum_value, chunk_size):
        start = i + 1  # Start from 1 for the first chunk
        end = min(i + chunk_size, max_sum_value)
        if start <= end:  # Ensure valid range
            ranges.append((start, end))
    
    # Use multiprocessing to find 2025-numbers in parallel
    with mp.Pool(processes=cpu_count) as pool:
        results = pool.starmap(find_2025_numbers_in_range, 
                              [(start, end, n) for start, end in ranges])
    
    # Combine and sort all results
    all_results = []
    for chunk_results in results:
        all_results.extend(chunk_results)
    
    all_results.sort()
    
    # Display results
    print(f"Found {len(all_results)} 2025-numbers with {n} digits or less")
    ### for num, a, b, sum_val in all_results:
    ###     print(f"{num} = concat({a}, {b}) = ({a} + {b})² = {sum_val}²")
    
    # Calculate and return the sum
    total = sum(num for num, _, _, _ in all_results)
    return total

def calculate_T(n):
    """Calculate T(n) - decide whether to use parallel processing based on n."""
    if n >= 10:
        print(f"Using parallel processing with {mp.cpu_count()} CPU cores...")
        return calculate_T_parallel(n)
    else:
        # For smaller n, use single-threaded approach
        results = find_2025_numbers_in_range(1, int(math.sqrt(10**n)) + 1, n)
        
        # Display results
        print(f"Found {len(results)} 2025-numbers with {n} digits or less")
        ### for num, a, b, sum_val in results:
        ###     print(f"{num} = concat({a}, {b}) = ({a} + {b})² = {sum_val}²")
        
        # Calculate and return the sum
        total = sum(num for num, _, _, _ in results)
        return total

def main():
    start_time = time.time()
    
    if len(sys.argv) > 1:
        n = int(sys.argv[1])
        result = calculate_T(n)
        print(f"T({n}) = {result}")
    else:
        # Verify that T(4) = 5131
        n = 4
        result = calculate_T(n)
        print(f"T({n}) = {result}")
        if result == 5131:
            print("Verification successful: T(4) = 5131")
        else:
            print(f"Verification failed: Expected T(4) = 5131, got {result}")
    
    end_time = time.time()
    print(f"Execution time: {end_time - start_time:.6f} seconds")

if __name__ == "__main__":
    main()
